"""
Some utility functions.
"""
from __future__ import annotations

import logging
import os
import sys
import uuid
from collections import deque
from datetime import datetime
from enum import Enum
from typing import TYPE_CHECKING

import numpy as np
import numpy.typing as npt
from scipy.spatial import distance_matrix

from overcooked_simulator import ROOT_DIR

if TYPE_CHECKING:
    from overcooked_simulator.counters import Counter


def create_init_env_time():
    """Init time of the environment time, because all environments should have the same internal time."""
    return datetime(
        year=2000, month=1, day=1, hour=0, minute=0, second=0, microsecond=0
    )


def get_closest(point: npt.NDArray[float], counters: list[Counter]):
    """Determines the closest counter for a given 2d-coordinate point in the env.

    Args:
        point: The point in the env for which to find the closest counter
        counters: List of objects with a `pos` attribute to compare to.

    Returns: The closest counter for the given point.
    """

    return counters[
        np.argmin(distance_matrix([point], [counter.pos for counter in counters])[0])
    ]


def get_touching_counters(target: Counter, counters: list[Counter]) -> list[Counter]:
    return list(
        filter(
            lambda counter: np.linalg.norm(counter.pos - target.pos) == 1.0, counters
        )
    )


def find_item_on_counters(item_uuid: str, counters: list[Counter]) -> Counter | None:
    for counter in counters:
        if counter.occupied_by:
            if isinstance(counter.occupied_by, deque):
                for item in counter.occupied_by:
                    if item.uuid == item_uuid:
                        return counter
            else:
                if item_uuid == counter.occupied_by.uuid:
                    return counter


def custom_asdict_factory(data):
    """Convert enums to their value."""

    def convert_value(obj):
        if isinstance(obj, Enum):
            return obj.value
        return obj

    return dict((k, convert_value(v)) for k, v in data)


def setup_logging(enable_websocket_logging=False):
    path_logs = ROOT_DIR.parent / "logs"
    os.makedirs(path_logs, exist_ok=True)
    logging.basicConfig(
        level=logging.DEBUG,
        format="%(asctime)s %(levelname)-8s %(name)-50s %(message)s",
        handlers=[
            logging.FileHandler(
                path_logs / f"{datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}_debug.log",
                encoding="utf-8",
            ),
            logging.StreamHandler(sys.stdout),
        ],
    )
    logging.getLogger("matplotlib").setLevel(logging.WARNING)
    if not enable_websocket_logging:
        logging.getLogger("asyncio").setLevel(logging.ERROR)
        logging.getLogger("asyncio.coroutines").setLevel(logging.ERROR)
        logging.getLogger("websockets.server").setLevel(logging.ERROR)
        logging.getLogger("websockets.protocol").setLevel(logging.ERROR)
        logging.getLogger("websockets.client").setLevel(logging.ERROR)


def url_and_port_arguments(parser):
    parser.add_argument(
        "-url",
        "--url",
        "--host",
        type=str,
        default="localhost",
        help="Overcooked game server host url.",
    )
    parser.add_argument(
        "-p",
        "--port",
        type=int,
        default=8000,
        help="Port number for the game engine server",
    )


def disable_websocket_logging_arguments(parser):
    parser.add_argument(
        "--enable-websocket-logging" "", action="store_true", default=True
    )


def add_list_of_manager_ids_arguments(parser):
    parser.add_argument(
        "-m",
        "--manager_ids",
        nargs="+",
        type=str,
        default=[uuid.uuid4().hex],
        help="List of manager IDs that can create environments.",
    )
